package ru.ulstu.extractor.rule.service;

import com.fuzzylite.Engine;
import com.fuzzylite.activation.Highest;
import com.fuzzylite.norm.t.AlgebraicProduct;
import com.fuzzylite.rule.Rule;
import com.fuzzylite.rule.RuleBlock;
import com.fuzzylite.term.Triangle;
import com.fuzzylite.variable.InputVariable;
import com.fuzzylite.variable.OutputVariable;
import org.springframework.stereotype.Service;
import ru.ulstu.extractor.rule.model.AntecedentValue;
import ru.ulstu.extractor.rule.model.DbRule;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

@Service
public class FuzzyInferenceService {
    private final static String RULE_TEMPLATE = "if %s is %s AND %s is %s then state is %s";
    private final RuleService ruleService;
    private final AntecedentValueService antecedentValueService;

    public FuzzyInferenceService(RuleService ruleService,
                                 AntecedentValueService antecedentValueService) {
        this.ruleService = ruleService;
        this.antecedentValueService = antecedentValueService;
    }

    public List<String> getRulesFromDb() {
        List<DbRule> dbDbRules = ruleService.getList();
        return dbDbRules.stream().map(this::getFuzzyRule).collect(Collectors.toList());
    }

    private String getFuzzyRule(DbRule dbRule) {
        return String.format(RULE_TEMPLATE,
                dbRule.getFirstAntecedent().name(),
                dbRule.getFirstAntecedentValue().getAntecedentValue(),
                dbRule.getSecondAntecedent().name(),
                dbRule.getSecondAntecedentValue().getAntecedentValue(),
                dbRule.getConsequent());
    }

    private RuleBlock getRuleBlock(Engine engine, Map<String, Double> variableValues, List<AntecedentValue> antecedentValues) {
        variableValues.forEach((key, value) -> {
            InputVariable input = new InputVariable();
            input.setName(key);
            input.setDescription("");
            input.setEnabled(true);
            input.setRange(0.000, 1.000);
            input.setLockValueInRange(false);
            for (int i = 0; i < antecedentValues.size(); i++) {
                input.addTerm(new Triangle(antecedentValues.get(i).getAntecedentValue(), i, i + 2));
            }
            engine.addInputVariable(input);
        });


        RuleBlock mamdani = new RuleBlock();
        mamdani.setName("mamdani");
        mamdani.setDescription("");
        mamdani.setEnabled(true);
        mamdani.setConjunction(null);
        mamdani.setDisjunction(null);
        mamdani.setImplication(new AlgebraicProduct());
        mamdani.setActivation(new Highest());
        getRulesFromDb().forEach(r -> mamdani.addRule(Rule.parse(r, engine)));
        return mamdani;
    }

    private Engine getFuzzyEngine() {
        Engine engine = new Engine();
        engine.setName("Git rules");
        engine.setDescription("");
        return engine;
    }

    public String run() {
        Engine engine = getFuzzyEngine();
        List<AntecedentValue> antecedentValues = antecedentValueService.getList();
        Map<String, Double> variableValues = new HashMap<>();
        engine.addRuleBlock(getRuleBlock(engine, variableValues, antecedentValues));
        return getConsequent(engine, variableValues);
    }

    private String getConsequent(Engine engine, Map<String, Double> variableValues) {
        OutputVariable outputVariable = engine.getOutputVariable("state");
        for (Map.Entry<String, Double> variableValue : variableValues.entrySet()) {
            InputVariable inputVariable = engine.getInputVariable(variableValue.getKey());
            inputVariable.setValue(variableValue.getValue());
        }
        engine.process();
        return outputVariable.highestMembership(outputVariable.getValue()).getSecond().getName();
    }
}
